import 'dart:async';

import 'package:msg/src/core/cmd.dart';
import 'package:msg/src/private/cmd_concrete.dart';
import 'package:msg/src/core/msg.dart';

class Store<Model> {
  final _msgController = StreamController<Msg<Model>>();
  final _cmdController = StreamController<Cmd<Model>>();
  final _modelController = StreamController<Model>.broadcast();

  late final StreamSubscription<void> _msgSubscription;
  late final StreamSubscription<void> _cmdSubscription;

  List<StreamSubscription<void>> _externalSubscriptions = [];

  Cmd<Model> Function(Model current)? _onDispose;

  Model _model;

  Store(
      {required Model initialModel,
      Cmd<Model> Function(Model initial)? initialCmd,
      List<FutureOr<Stream<Msg<Model>>>> Function(Model initial)? subscriptions,
      Cmd<Model> Function(Model current)? onDispose})
      : _model = initialModel,
        _onDispose = onDispose {
    _msgSubscription = _msgController.stream.listen((msg) {
      final old = _model;
      if (msg.update != null) {
        _model = msg.update!(_model);
        if (_model != old) {
          _modelController.add(_model);
        }
      }
      if (msg.cmd != null) {
        _cmdController.add(msg.cmd!(old));
      }
    });

    _cmdSubscription = _cmdController.stream.listen((cmd) async {
      if (cmd is CmdMsg<Model>) _msgController.add(await cmd.msg);
      if (cmd is CmdEffect<Model>) cmd.effect();
      if (cmd is CmdBatch<Model>) cmd.cmds.forEach(_cmdController.add);
      if (cmd is CmdStream<Model>) cmd.cmds.forEach(_cmdController.add);
    });

    if (subscriptions != null) {
      _wireSubscriptions(subscriptions(initialModel));
    }

    if (initialCmd != null) {
      _cmdController.add(initialCmd(initialModel));
    }
  }

  Future<void> _wireSubscriptions(
      List<FutureOr<Stream<Msg<Model>>>> streams) async {
    for (final stream in streams) {
      _externalSubscriptions.add((await stream).listen(_msgController.add));
    }
  }

  void send(Msg<Model> msg) => _msgController.add(msg);

  Future<void> dispose() async {
    if (_onDispose != null) {
      _cmdController.add(_onDispose!(_model));
      await Future.delayed(Duration.zero);
    }
    await _msgSubscription.cancel();
    await _cmdSubscription.cancel();
    for (final subscription in _externalSubscriptions) {
      await subscription.cancel();
    }
    await _cmdController.close();
    await _msgController.close();
    await _modelController.close();
  }

  Model get model => _model;

  Stream<Model> get stream => _modelController.stream;
}
